# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is The Browser Profile Migrator.
#
# The Initial Developer of the Original Code is Ben Goodger.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ben Goodger <ben@bengoodger.com>
#   Mark Banner <bugzilla@standard8.demon.co.uk>
#   Postbox, Inc. <contact@postbox-inc.com>
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

const nsIMailProfileMigrator = Components.interfaces.nsIMailProfileMigrator;
const nsIProfileStartup        = Components.interfaces.nsIProfileStartup;
const nsISupportsString       = Components.interfaces.nsISupportsString;
const NS_PROFILE_MIGRATOR_CONTRACTID = "@mozilla.org/profile/migrator;1?app=mail&type=";

var MigrationWizard = {
  _source: "",                  // Source Profile Migrator ContractID suffix
  _itemsFlags: nsIMailProfileMigrator.ALL,       // Selected Import Data Sources (16-bit bitfield)
  _selectedProfile: null,       // Selected Profile name to import from
  _wiz: null,
  _migrator: null,
  _singleItem: false,           // Are we choosing just to import a single
                                // item into the current profile?
  _autoMigrate: null,

  init: function ()
  {
    var os = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    os.addObserver(this, "Migration:Started", false);
    os.addObserver(this, "Migration:ItemBeforeMigrate", false);
    os.addObserver(this, "Migration:ItemAfterMigrate", false);
    os.addObserver(this, "Migration:Ended", false);
    os.addObserver(this, "Migration:Progress", false);

    this._wiz = document.documentElement;
    this._wiz.canRewind = false;

    if ("arguments" in window) {
      this._source = window.arguments[0];
      this._migrator = window.arguments[1] ? window.arguments[1].QueryInterface(nsIMailProfileMigrator) : null;
      this._autoMigrate = window.arguments[2] ? window.arguments[2].QueryInterface(nsIProfileStartup) : null;
    }

    this.onWelcomePageShow();
  },

  uninit: function ()
  {
    var os = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    os.removeObserver(this, "Migration:Started");
    os.removeObserver(this, "Migration:ItemBeforeMigrate");
    os.removeObserver(this, "Migration:ItemAfterMigrate");
    os.removeObserver(this, "Migration:Ended");
    os.removeObserver(this, "Migration:Progress");
  },

  // 0 - Welcome
  onWelcomePageShow: function()
  {
  },
  
  onVisitQuickstartGuide: function()
  {
    var quickstartGuideURL = "http://www.postbox-inc.com/quickstart";
    var uri = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService).newURI(quickstartGuideURL, null, null);
    var protocolSvc = Components.classes["@mozilla.org/uriloader/external-protocol-service;1"].
    getService(Components.interfaces.nsIExternalProtocolService);
    protocolSvc.loadUrl(uri);
  },
  
  onStart: function()
  {
    this._wiz.canAdvance = true;
    
    // Figure out what source apps are are available to import from
    this.onImportSourcePageShow();
    // if no apps are available, make "Start Using Postbox" cancel out
    var group = document.getElementById("importSourceGroup");
    var availableApps = 0;
    for (var i = 0; i < group.childNodes.length; i++) {
      if (group.childNodes[i].hidden == false)
        availableApps++;
    }
    
    if (availableApps)
    {
      if (availableApps == 1)
      {
        group.setAttribute("noradioimages", "true");
      }
      this._wiz.advance();
    }
    else
    {
      document.documentElement.cancel();
    }
  },
  
  // 1 - Import Source
  onImportSourcePageShow: function ()
  {
    // Figure out what source apps are are available to import from:
    var group = document.getElementById("importSourceGroup");
    var firstSelectable = null;
    for (var i = 0; i < group.childNodes.length; ++i) {
      var suffix = group.childNodes[i].id;
      var contractID = NS_PROFILE_MIGRATOR_CONTRACTID + suffix;
      var migrator = null;
      if (contractID in Components.classes) {
        migrator = Components.classes[contractID]
                             .createInstance(nsIMailProfileMigrator);
      } else {
        dump("*** invalid contractID =" + contractID + "\n");
        // This is an invalid contract id, therefore hide this element
        // and allow things to continue - that way we should be able to
        // copy with anything happening.
        group.childNodes[i].hidden = true;
        break;
      }
      // Ensure that we only allow import selections for profile
      // migrators that support the requested action.
      if (!migrator.sourceExists || !(migrator.supportedItems & this._itemsFlags)) {
        group.childNodes[i].hidden = true;
        continue;
      }
      if (!firstSelectable && !group.childNodes[i].disabled &&
          !group.childNodes[i].hidden) {
        firstSelectable = group.childNodes[i];
      }
    }

    if (this._source) {
      // Somehow the Profile Migrator got confused, and gave us a migrate source
      // that doesn't actually exist. This could be because of a bogus registry
      // state. Set the _source property to null so the first visible item in
      // the list is selected instead.
      if (document.getElementById(this._source).hidden)
        this._source = null;
    }
    group.selectedItem = this._source ?
      document.getElementById(this._source) : firstSelectable;
  },

  onWelcomePageAdvanced: function()
  {
    return true;
  },
  
  onImportSourcePageAdvanced: function ()
  {
    let importSourceGroupSelItem = document.getElementById("importSourceGroup").selectedItem;
    var newSource = importSourceGroupSelItem.id;

    if (!this._migrator || newSource != this._source) {
      // Create the migrator for the selected source.
      var contractID = NS_PROFILE_MIGRATOR_CONTRACTID + newSource;
      this._migrator = Components.classes[contractID]
                                 .createInstance(nsIMailProfileMigrator);
      this._selectedProfile = null;
    }
    this._source = newSource;

    // check for more than one source profile
    if (this._migrator.sourceHasMultipleProfiles)
      this._wiz.currentPage.next = "selectProfile";
    else {
      if (this._autoMigrate)
        this._wiz.currentPage.next = "migrating";
      else if (this._singleItem)
        this._wiz.currentPage.next = "migrating";
      else
        this._wiz.currentPage.next = "importItems";

      var sourceProfiles = this._migrator.sourceProfiles;
      if (sourceProfiles && sourceProfiles.length == 1) {
        this._selectedProfile = sourceProfiles
            .queryElementAt(0, nsISupportsString).data;
      }
      else
        this._selectedProfile = "";
    }
    return true;
  },

  // 2 - [Profile Selection]
  onSelectProfilePageShow: function() {
    var profiles = document.getElementById("profiles");
    while (profiles.hasChildNodes())
      profiles.removeChild(profiles.lastChild);

    var sourceProfiles = this._migrator.sourceProfiles;
    var count = sourceProfiles.length;
    for (var i = 0; i < count; ++i) {
      var item = document.createElement("radio");
      item.id = sourceProfiles.queryElementAt(i, nsISupportsString).data;
      item.setAttribute("label", item.id);
      profiles.appendChild(item);
    }

    profiles.selectedItem = this._selectedProfile ?
      document.getElementById(this._selectedProfile) : profiles.firstChild;
  },

  onSelectProfilePageAdvanced: function() {
    this._selectedProfile = document.getElementById("profiles").selectedItem.id;

    // If we're automigrating don't show the item
    // selection page
    if (this._autoMigrate)
      this._wiz.currentPage.next = "migrating"
  },

  // 3 - ImportItems
  onImportItemsPageShow: function ()
  {
    var dataSources = document.getElementById("dataSources");
    while (dataSources.hasChildNodes())
      dataSources.removeChild(dataSources.firstChild);

    var bundle = document.getElementById("bundle");

    var items = this._migrator.getMigrateData(this._selectedProfile, this._autoMigrate);
    for (var i = 0; i < 16; ++i) {
      var itemID = items & (1 << i);
      if (itemID) {
        var checkbox = document.createElement("checkbox");
        checkbox.id = itemID;
        try {
          checkbox.setAttribute("label",
                                bundle.getString(itemID + "_" + this._source));
        }
        catch (ex) {
          checkbox.setAttribute("label",
                                bundle.getString(itemID + "_generic"));
        }
        dataSources.appendChild(checkbox);
        if (this._itemsFlags & itemID)
          checkbox.setAttribute("checked", true);
      }
    }
  },

  onImportItemsPageRewound: function() {
    this._wiz.canAdvance = true;
    this.onImportItemsPageAdvanced();
  },

  onImportItemsPageAdvanced: function() {
    var dataSources = document.getElementById("dataSources");
    this._itemsFlags = 0;
    for (var i = 0; i < dataSources.childNodes.length; ++i) {
      var checkbox = dataSources.childNodes[i];
      if (checkbox.checked)
      {
        // we are hiding "Save Passwords" so import them if
        // we are importing "Account Settings"
        if (checkbox.id != Components.interfaces.nsIMailProfileMigrator.PASSWORDS ||
            document.getElementById(Components.interfaces.nsIMailProfileMigrator.ACCOUNT_SETTINGS).checked)
        {
          this._itemsFlags |= checkbox.id;
        }
      }
    }
  },

  onImportItemCommand: function(aEvent) {
    this._wiz.canAdvance = document
      .getElementById("dataSources")
      .getElementsByAttribute("checked", "true").item(0) != null;
  },

  // 4 - Migrating
  onMigratingPageShow: function ()
  {
    this._wiz.getButton("cancel").disabled = true;
    this._wiz.canRewind = false;
    this._wiz.canAdvance = false;

    // When automigrating, show all of the data that can be received from this source.
    if (this._autoMigrate)
      this._itemsFlags = this._migrator.getMigrateData(this._selectedProfile, this._autoMigrate);

    this._listItems("migratingItems");
    setTimeout(this.onMigratingMigrate, 0, this);
  },

  onMigratingMigrate: function (aOuter)
  {
    aOuter._migrator.migrate(aOuter._itemsFlags, aOuter._autoMigrate, aOuter._selectedProfile);
  },

  _listItems: function(aID) {
    var items = document.getElementById(aID);
    while (items.hasChildNodes())
      items.removeChild(items.lastChild);

    var bundle = document.getElementById("bundle");
    var itemID;
    for (var x = 1; x < nsIMailProfileMigrator.ALL;
         x = x << 1) {
      if (x & this._itemsFlags) {
        var label = document.createElement("label");
        label.id = x + "_migrated";
        
        // skip certain things so everything fits in the wizard
        // see bug #7689 for details
        if (x == Components.interfaces.nsIMailProfileMigrator.SETTINGS ||
            x == Components.interfaces.nsIMailProfileMigrator.PASSWORDS ||
            x == Components.interfaces.nsIMailProfileMigrator.JUNKTRAINING ||
            x == Components.interfaces.nsIMailProfileMigrator.OTHERDATA)
        {
          continue;
        }
        
        try {
          label.setAttribute("value",
                             bundle.stringBundle
                                   .GetStringFromName(x + "_"+ this._source));
          label.setAttribute("class", "migration-pending");
          items.appendChild(label);
        }
        catch (ex) {
          try {
            label.setAttribute("value", bundle.getString(x + "_generic"));
            label.setAttribute("class", "migration-pending");
            items.appendChild(label);
          }
          catch (e) {
            // if the block above throws, we've enumerated all the import
            // data types we currently support and are now just wasting time.
            break;
          }
        }
      }
    }
  },

  observe: function(aSubject, aTopic, aData) {
    switch (aTopic) {
    case "Migration:Progress":
      document.getElementById("progressBar").value = aData;
      break;
    case "Migration:Started":
      break;
    case "Migration:ItemBeforeMigrate":
      var label = document.getElementById(aData + "_migrated");
      if (label)
        label.setAttribute("class", "migration-in-progress");
      break;
    case "Migration:ItemAfterMigrate":
      var label = document.getElementById(aData + "_migrated");
      if (label)
        label.setAttribute("class", "migration-finished");
      break;
    case "Migration:Ended":
      // We're done now.
      this._wiz.canAdvance = true;
      this._wiz.advance();

      if (this._autoMigrate)
        setTimeout(function() {window.close();}, 5000);

      break;
    }
  },

  onDonePageShow: function ()
  {
    this._wiz.getButton("cancel").disabled = true;
    this._wiz.canRewind = false;
    this._listItems("doneItems");
  }
};

function pbOnWizardPageClick(aEvent)
{
  if (aEvent.originalTarget.className == 'wizard-header-icon')
    openRegionURL('mailnews.hints_and_tips.url');
}
