/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is The Browser Profile Migrator.
 *
 * The Initial Developer of the Original Code is Ben Goodger.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Ben Goodger <ben@bengoodger.com>
 *   Postbox, Inc. <contact@postbox-inc.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsAppDirectoryServiceDefs.h"
#include "nsMailProfileMigratorUtils.h"
#include "nsCRT.h"
#include "nsDirectoryServiceDefs.h"
#include "nsIObserverService.h"
#include "nsIPrefLocalizedString.h"
#include "nsIPrefService.h"
#include "nsIServiceManager.h"
#include "nsISupportsArray.h"
#include "nsISupportsPrimitives.h"
#include "nsNetCID.h"
#include "nsNetUtil.h"
#include "nsThunderbirdProfileMigrator.h"
#include "nsINIParser.h"
#include "prprf.h"

///////////////////////////////////////////////////////////////////////////////
// nsThunderbirdProfileMigrator

NS_IMPL_ISUPPORTS2(nsThunderbirdProfileMigrator, nsIMailProfileMigrator,
                   nsITimerCallback)

nsThunderbirdProfileMigrator::nsThunderbirdProfileMigrator()
{
}

nsThunderbirdProfileMigrator::~nsThunderbirdProfileMigrator()
{
}

///////////////////////////////////////////////////////////////////////////////
// nsIMailProfileMigrator

NS_IMETHODIMP
nsThunderbirdProfileMigrator::Migrate(PRUint16 aItems,
                                      nsIProfileStartup* aStartup,
                                      const PRUnichar* aProfile)
{
  nsresult rv = NS_OK;
  PRBool aReplace = aStartup ? PR_TRUE : PR_FALSE;

  if (!mTargetProfile) {
    GetProfilePath(aStartup, getter_AddRefs(mTargetProfile));
    if (!mTargetProfile) return NS_ERROR_FAILURE;
  }
  if (!mSourceProfile)
    GetSourceProfile(aProfile);

  NOTIFY_OBSERVERS(MIGRATION_STARTED, nsnull);

  COPY_DATA(CopyPreferences,  aReplace, nsIMailProfileMigrator::SETTINGS);
  COPY_DATA(CopyPasswords,    aReplace, nsIMailProfileMigrator::PASSWORDS);
  COPY_DATA(CopyOtherData,    aReplace, nsIMailProfileMigrator::OTHERDATA);


  // fake notifications for things we've already imported as part of
  // CopyPreferences
  nsAutoString index;
  index.AppendInt(nsIMailProfileMigrator::ACCOUNT_SETTINGS);
  NOTIFY_OBSERVERS(MIGRATION_ITEMBEFOREMIGRATE, index.get());
  NOTIFY_OBSERVERS(MIGRATION_ITEMAFTERMIGRATE, index.get());

  index.Truncate();
  index.AppendInt(nsIMailProfileMigrator::NEWSDATA);
  NOTIFY_OBSERVERS(MIGRATION_ITEMBEFOREMIGRATE, index.get());
  NOTIFY_OBSERVERS(MIGRATION_ITEMAFTERMIGRATE, index.get());

  // copy junk mail training file
  COPY_DATA(CopyJunkTraining, aReplace, nsIMailProfileMigrator::JUNKTRAINING);

  if (aReplace &&
      (aItems & nsIMailProfileMigrator::SETTINGS ||
       aItems & nsIMailProfileMigrator::PASSWORDS ||
       !aItems)) {
    // Permissions (Images, Cookies, Popups)
    rv |= CopyFile(FILE_NAME_SITEPERM_NEW, FILE_NAME_SITEPERM_NEW);
    rv |= CopyFile(FILE_NAME_SITEPERM_OLD, FILE_NAME_SITEPERM_OLD);
  }

  // the last thing to do is to actually copy over any mail folders
  // we have marked for copying we want to do this last and it will be
  // asynchronous so the UI doesn't freeze up while we perform
  // this potentially very long operation.
  CopyMailFolders();

  return rv;
}

NS_IMETHODIMP
nsThunderbirdProfileMigrator::GetMigrateData(const PRUnichar* aProfile,
                                             PRBool aReplace,
                                             PRUint16* aResult)
{
  *aResult = 0;

  if (!mSourceProfile) {
    GetSourceProfile(aProfile);
    if (!mSourceProfile)
      return NS_ERROR_FILE_NOT_FOUND;
  }

  // migration fields for things we always migrate
  *aResult =
    nsIMailProfileMigrator::ACCOUNT_SETTINGS |
    nsIMailProfileMigrator::MAILDATA |
    nsIMailProfileMigrator::NEWSDATA |
    nsIMailProfileMigrator::ADDRESSBOOK_DATA |
    nsIMailProfileMigrator::PASSWORDS;

  MigrationData data[] = { { FILE_NAME_PREFS,
                             nsIMailProfileMigrator::SETTINGS,
                             PR_TRUE },
                           { FILE_NAME_USER_PREFS,
                             nsIMailProfileMigrator::SETTINGS,
                             PR_TRUE },
                           { FILE_NAME_MIMETYPES,
                             nsIMailProfileMigrator::OTHERDATA,
                             PR_TRUE },
                           { FILE_NAME_JUNKTRAINING,
                             nsIMailProfileMigrator::JUNKTRAINING,
                             PR_TRUE } };
                                                                  
  GetMigrateDataFromArray(data, sizeof(data)/sizeof(MigrationData),
                          aReplace, mSourceProfile, aResult);

  return NS_OK;
}

NS_IMETHODIMP
nsThunderbirdProfileMigrator::GetSupportedItems(PRUint16 *aSupportedItems)
{
  NS_ENSURE_ARG_POINTER(aSupportedItems);

  *aSupportedItems = nsIMailProfileMigrator::SETTINGS |
    nsIMailProfileMigrator::OTHERDATA |
    nsIMailProfileMigrator::JUNKTRAINING |
    nsIMailProfileMigrator::PASSWORDS |
    nsIMailProfileMigrator::ACCOUNT_SETTINGS |
    nsIMailProfileMigrator::MAILDATA |
    nsIMailProfileMigrator::NEWSDATA |
    nsIMailProfileMigrator::ADDRESSBOOK_DATA;

  return NS_OK;
}

///////////////////////////////////////////////////////////////////////////////
// nsThunderbirdProfileMigrator

nsresult
nsThunderbirdProfileMigrator::FillProfileDataFromRegistry()
{
  // Find the Thunderbird Registry
  nsCOMPtr<nsIProperties> fileLocator(
    do_GetService("@mozilla.org/file/directory_service;1"));
  nsCOMPtr<nsILocalFile> thunderbirdData;
#ifdef XP_WIN
#define REGISTRY_FILE "registry.dat"
  fileLocator->Get(NS_WIN_APPDATA_DIR, NS_GET_IID(nsILocalFile),
                   getter_AddRefs(thunderbirdData));

  thunderbirdData->Append(NS_LITERAL_STRING("Thunderbird"));

#elif defined(XP_MACOSX)
#define REGISTRY_FILE "Application Registry"
  fileLocator->Get(NS_MAC_USER_LIB_DIR, NS_GET_IID(nsILocalFile),
                   getter_AddRefs(thunderbirdData));
  
  thunderbirdData->Append(NS_LITERAL_STRING("Thunderbird"));

#elif defined(XP_UNIX)
#define REGISTRY_FILE "appreg"
  fileLocator->Get(NS_UNIX_HOME_DIR, NS_GET_IID(nsILocalFile),
                   getter_AddRefs(thunderbirdData));
  
  thunderbirdData->Append(NS_LITERAL_STRING(".thunderbird"));

#elif defined(XP_BEOS)
#define REGISTRY_FILE "appreg"
   fileLocator->Get(NS_BEOS_SETTINGS_DIR, NS_GET_IID(nsILocalFile),
                    getter_AddRefs(thunderbirdData));

   thunderbirdData->Append(NS_LITERAL_STRING("Thunderbird"));

#elif defined(XP_OS2)
#define REGISTRY_FILE "registry.dat"
  fileLocator->Get(NS_OS2_HOME_DIR, NS_GET_IID(nsILocalFile),
                   getter_AddRefs(thunderbirdData));
  
  thunderbirdData->Append(NS_LITERAL_STRING("Thunderbird"));
#else
  // On other OS just abort
  return NS_ERROR_FAILURE;
#endif

  // Try profiles.ini first
  nsresult rv = GetProfileDataFromProfilesIni(thunderbirdData,
                                              mProfileNames,
                                              mProfileLocations);

  if (rv != NS_ERROR_FILE_NOT_FOUND)
    return rv;

  thunderbirdData->Append(NS_LITERAL_STRING(REGISTRY_FILE));

  // Then try the old registry format
  return GetProfileDataFromRegistry(thunderbirdData, mProfileNames,
                                    mProfileLocations);
}

static
nsThunderbirdProfileMigrator::PrefTransform gTransforms[] = {

  MAKESAMETYPEPREFTRANSFORM("mail.accountmanager.accounts",            String),
  MAKESAMETYPEPREFTRANSFORM("mail.accountmanager.defaultaccount",      String),
  MAKESAMETYPEPREFTRANSFORM("mail.accountmanager.localfoldersserver",  String),
  MAKESAMETYPEPREFTRANSFORM("mail.accountwizard.deferstorage",         Bool),
  MAKESAMETYPEPREFTRANSFORM("mail.collect_addressbook",                String),

  MAKESAMETYPEPREFTRANSFORM("mail.default_sendlater_uri",              String),
  MAKESAMETYPEPREFTRANSFORM("mail.delete_matches_sort_order",          Bool),
  MAKESAMETYPEPREFTRANSFORM("mail.fcc_folder",                         String),

  MAKESAMETYPEPREFTRANSFORM("mail.pop3.deleteFromServerOnMove",        Bool),
  MAKESAMETYPEPREFTRANSFORM("mail.prompt_purge_threshhold",            Bool),

  MAKESAMETYPEPREFTRANSFORM("mail.smtp.defaultserver",                 String),
  MAKESAMETYPEPREFTRANSFORM("mail.smtpservers",                        String),

  MAKESAMETYPEPREFTRANSFORM("mailnews.customDBHeaders",                String),
  MAKESAMETYPEPREFTRANSFORM("mailnews.customHeaders",                  String),

  MAKESAMETYPEPREFTRANSFORM("mailnews.html_domains",                   String),

  MAKESAMETYPEPREFTRANSFORM("mailnews.plaintext_domains",              String),
  MAKESAMETYPEPREFTRANSFORM("mailnews.ui.junk.firstuse",               Bool),

  MAKESAMETYPEPREFTRANSFORM("security.default_personal_cert",          String),

  MAKESAMETYPEPREFTRANSFORM("signon.SignonFileName",                   String),
  MAKESAMETYPEPREFTRANSFORM("signon.expireMasterPassword",             Bool),
};

nsresult
nsThunderbirdProfileMigrator::TransformPreferences(
  const char* aSourcePrefFileName,
  const char* aTargetPrefFileName)
{
  PrefTransform* transform;
  PrefTransform* end = gTransforms + sizeof(gTransforms)/sizeof(PrefTransform);

  // Load the source pref file
  nsCOMPtr<nsIPrefService> psvc(do_GetService(NS_PREFSERVICE_CONTRACTID));
  psvc->ResetPrefs();

  nsCOMPtr<nsIFile> sourcePrefsFile;
  mSourceProfile->Clone(getter_AddRefs(sourcePrefsFile));
  sourcePrefsFile->AppendNative(nsDependentCString(aSourcePrefFileName));
  psvc->ReadUserPrefs(sourcePrefsFile);

  nsCOMPtr<nsIPrefBranch> branch(do_QueryInterface(psvc));
  for (transform = gTransforms; transform < end; ++transform)
    transform->prefGetterFunc(transform, branch);

  // read in the various pref branch trees for accounts, identities, servers,
  // etc.
  static const char* branchNames[] =
  {
    // Keep the three below first, or change the indexes below
    "mail.identity.",
    "mail.server.",
    "ldap_2.",
    "mail.account.",
    "mail.addr_book.",
    "mail.imap.",
    "mail.smtpserver.",
    "mailnews.tags.",
    "privacy.",
    "security.OSCP.",
    "security.crl.",
    "wallet."
  };

  PBStructArray branches[NS_ARRAY_LENGTH(branchNames)];
  PRUint32 i;
  for (i = 0; i < NS_ARRAY_LENGTH(branchNames); ++i)
    ReadBranch(branchNames[i], psvc, branches[i]);

  // certain mail prefs may actually be absolute paths instead of profile
  // relative paths we need to fix these paths up before we write them out to
  // the new prefs.js
  CopyMailFolderPrefs(branches[1], psvc);

  CopyAddressBookDirectories(branches[2], psvc);

  // Now that we have all the pref data in memory, load the target pref file,
  // and write it back out
  psvc->ResetPrefs();

  nsCOMPtr<nsIFile> targetPrefsFile;
  mTargetProfile->Clone(getter_AddRefs(targetPrefsFile));
  targetPrefsFile->AppendNative(nsDependentCString(aTargetPrefFileName));

  // Don't use nsnull here as we're too early in the cycle for the prefs
  // service to get its default file (because the NS_GetDirectoryService items
  // aren't fully set up yet).
  psvc->ReadUserPrefs(targetPrefsFile);
  for (transform = gTransforms; transform < end; ++transform)
    transform->prefSetterFunc(transform, branch);

  for (i = 0; i < NS_ARRAY_LENGTH(branchNames); ++i)
    WriteBranch(branchNames[i], psvc, branches[i]);

  nsCOMPtr<nsIPrefBranch> prefBranch;
  psvc->GetBranch (nsnull, getter_AddRefs(prefBranch));
  if (prefBranch)
    prefBranch->SetCharPref("pb.migrated_source", "thunderbird");
  psvc->SavePrefFile(targetPrefsFile);

  return NS_OK;
}

nsresult
nsThunderbirdProfileMigrator::CopyLightningData()
{
#if 0
  // copy the Lightning databse (which has the settings)
  // but only if we are coming from Thunderbird 2.
  // the settings for Lightning 1.0 pre / Thunderbird 3 are not
  // compatible with Lightning 0.9

  nsCOMPtr<nsIFile> compatibilityFile;
  mSourceProfile->Clone(getter_AddRefs(compatibilityFile));

  compatibilityFile->AppendNative(nsDependentCString("compatibility.ini"));
  PRBool exists = PR_FALSE;
  compatibilityFile->Exists(&exists);
  if (!exists)
    return NS_OK;

  nsCOMPtr<nsILocalFile> localFile = do_QueryInterface(compatibilityFile);

  nsINIParser parser;
  nsresult rv = parser.Init(localFile);
  // Init does not fail on parsing errors, only on OOM/really unexpected
  // conditions.
  NS_ENSURE_SUCCESS(rv, rv);

  nsCAutoString buffer;
  rv = parser.GetString("Compatibility", "LastVersion", buffer);
  if (NS_SUCCEEDED(rv) && Substring(buffer, 0, 2).EqualsLiteral("2."))
  {
    (void)CopyFile(FILE_NAME_LIGHTNING_DB, FILE_NAME_LIGHTNING_DB);
  }
#endif
  return NS_OK;
}

nsresult
nsThunderbirdProfileMigrator::CopyPreferences(PRBool aReplace)
{
  nsresult rv = NS_OK;
  if (!aReplace)
    return rv;

  rv |= TransformPreferences(FILE_NAME_PREFS, FILE_NAME_PREFS);
  rv |= CopyFile(FILE_NAME_USER_PREFS, FILE_NAME_USER_PREFS);

  // Security Stuff
  rv |= CopyFile(FILE_NAME_CERT8DB, FILE_NAME_CERT8DB);
  rv |= CopyFile(FILE_NAME_KEY3DB, FILE_NAME_KEY3DB);
  rv |= CopyFile(FILE_NAME_SECMODDB, FILE_NAME_SECMODDB);
  rv |= CopyFile(FILE_NAME_CERTOVERRIDE, FILE_NAME_CERTOVERRIDE);

  // User MIME Type overrides

  return NS_OK;
}

nsresult
nsThunderbirdProfileMigrator::CopyOtherData(PRBool aReplace)
{
  return NS_OK;
}

/**
 * when migrating from Thunderbird, we rely on the relative directoryprefs so we need to subclass this code from nsNetscapeProfileMigratorBase
 * which works with the absolute directory paths common in other, older mail clients (netscape 6, mozilla suite, etc)
 *
 */
nsresult
nsThunderbirdProfileMigrator::CopyMailFolderPrefs(PBStructArray &aMailServers,
                                                   nsIPrefService* aPrefService)
{
  // Each server has a .directory pref which points to the location of the
  // mail data for that server. We need to do two things for that case...
  // (1) Fix up the directory path for the new profile
  // (2) copy the mail folder data from the source directory pref to the
  //     destination directory pref

  PRInt32 count = aMailServers.Length();
  for (PRInt32 i = 0; i < count; ++i) {
    PrefBranchStruct* pref = aMailServers.ElementAt(i);
    nsDependentCString prefName(pref->prefName);
    if (StringEndsWith(prefName, NS_LITERAL_CSTRING(".directory")) ||
        StringEndsWith(prefName, NS_LITERAL_CSTRING(".newsrc-file"))) {
      // we rely on the relative paths
      if (pref->type == nsIPrefBranch::PREF_STRING)
        NS_Free(pref->stringValue);

      aMailServers.RemoveElementAt(i);
      // Now decrease i and count to match the removed element
      --i;
      --count;
    }

    if (StringEndsWith(prefName, NS_LITERAL_CSTRING(".directory-rel"))) {
      // let's try to get a branch for this particular server to simplify things
      prefName.Cut(prefName.Length() - strlen("directory-rel"), strlen("directory-rel"));
      prefName.Insert("mail.server.", 0);

      nsCOMPtr<nsIPrefBranch> serverBranch;
      aPrefService->GetBranch(prefName.get(), getter_AddRefs(serverBranch));

      if (!serverBranch)
        break; // should we clear out this server pref from aMailServers?

      // convert the relative pref into a full path
      nsCString utf8RelativePath;
      nsresult rv = serverBranch->GetCharPref("directory-rel", getter_Copies(utf8RelativePath));
      if (NS_FAILED(rv))
        break;

      nsCOMPtr<nsILocalFile> sourceMailFolder;
      rv = ConvertFromRelativePref(utf8RelativePath, mSourceProfile, getter_AddRefs(sourceMailFolder));
      if (NS_FAILED(rv) || !sourceMailFolder)
        break;

      nsCString serverType;
      serverBranch->GetCharPref("type", getter_Copies(serverType));

      // now based on type, we need to build a new destination path for the
      // mail folders for this server
      PRBool isImap = PR_FALSE;
      nsCOMPtr<nsIFile> targetMailFolder;
      if (serverType.Equals("imap")) {
        mTargetProfile->Clone(getter_AddRefs(targetMailFolder));
        targetMailFolder->Append(IMAP_MAIL_DIR_50_NAME);
        isImap = PR_TRUE;
        continue; // skip this 
      }
      else if (serverType.Equals("none") || serverType.Equals("pop3") || serverType.Equals("rss") ||
               serverType.Equals("movemail")) {
        // local folders and POP3 servers go under <profile>\Mail
        mTargetProfile->Clone(getter_AddRefs(targetMailFolder));
        targetMailFolder->Append(MAIL_DIR_50_NAME);
      }
      else if (serverType.Equals("nntp")) {
        mTargetProfile->Clone(getter_AddRefs(targetMailFolder));
        targetMailFolder->Append(NEWS_DIR_50_NAME);
      }

      if (targetMailFolder) {
        // for all of our server types, append the host name to the directory
        // as part of the new location
        nsCString hostName;
        serverBranch->GetCharPref("hostname", getter_Copies(hostName));
        targetMailFolder->Append(NS_ConvertASCIItoUTF16(hostName));

        // we should make sure the host name based directory we are going to
        // migrate the accounts into is unique. This protects against the
        // case where the user has multiple servers with the same host name.
        targetMailFolder->CreateUnique(nsIFile::DIRECTORY_TYPE, 0777);
        
        // if this is an imap account, the only file we want to migrate is the filter file (if there is one).
        // no need to recursively copy the mail folders. See Bug 3939 for details.
        if (isImap)
        {
        }
        else
          RecursiveCopy(sourceMailFolder, targetMailFolder);

        // now we want to make sure the actual relative directory pref that gets
        // transformed into the new profile's pref.js has the right file
        // location.
        nsCAutoString descriptorString;
        nsCOMPtr<nsILocalFile> localFile = do_QueryInterface(targetMailFolder);
        nsCOMPtr<nsILocalFile> targetProfile = do_QueryInterface(mTargetProfile);
        ConvertToRelativePref(localFile, targetProfile, descriptorString);
        NS_Free(pref->stringValue);
        pref->stringValue = ToNewCString(descriptorString);
      }
    }
    else if (StringEndsWith(prefName, NS_LITERAL_CSTRING(".newsrc.file-rel"))) {
      // copy the news RC file into \News. this won't work if the user has
      // different newsrc files for each account I don't know what to do in
      // that situation.
      nsCOMPtr<nsIFile> targetNewsRCFile;
      mTargetProfile->Clone(getter_AddRefs(targetNewsRCFile));
      targetNewsRCFile->Append(NEWS_DIR_50_NAME);

      // turn the pref into a nsILocalFile
      nsCOMPtr<nsILocalFile> srcNewsRCFile;
      nsresult rv = ConvertFromRelativePref(nsDependentCString(pref->stringValue), mSourceProfile, getter_AddRefs(srcNewsRCFile));
      if (NS_FAILED(rv) || !srcNewsRCFile)
        break;

      // now make the copy
      PRBool exists;
      srcNewsRCFile->Exists(&exists);
      if (exists) {
        nsAutoString leafName;
        srcNewsRCFile->GetLeafName(leafName);
        // will fail if we've already copied a newsrc file here
        srcNewsRCFile->CopyTo(targetNewsRCFile,leafName);
        targetNewsRCFile->Append(leafName);

        // now write out the new descriptor
        nsCAutoString descriptorString;
        nsCOMPtr<nsILocalFile> localFile = do_QueryInterface(targetNewsRCFile);
        nsCOMPtr<nsILocalFile> targetProfile = do_QueryInterface(mTargetProfile);
        ConvertToRelativePref(localFile, targetProfile, descriptorString);
        NS_Free(pref->stringValue);
        pref->stringValue = ToNewCString(descriptorString);
      }
    }
  }

  return NS_OK;
}

/**
 * when migrating from Thunderbird, we rely on the relative directoryprefs so we need to subclass this code from nsNetscapeProfileMigratorBase
 * which works with the absolute directory paths common in other, older mail clients (netscape 6, mozilla suite, etc)
 *
 */
nsresult
nsThunderbirdProfileMigrator::CopySignatureFiles(PBStructArray &aIdentities,
                                                  nsIPrefService* aPrefService)
{
  nsresult rv = NS_OK;

  PRUint32 count = aIdentities.Length();
  for (PRUint32 i = 0; i < count; ++i)
  {
    PrefBranchStruct* pref = aIdentities.ElementAt(i);
    nsDependentCString prefName(pref->prefName);

    if (StringEndsWith(prefName, NS_LITERAL_CSTRING(".sig_file-rel")))
    {
      // turn the pref into a nsILocalFile
      nsCOMPtr<nsILocalFile> srcSignatureFile;
      rv = ConvertFromRelativePref(nsDependentCString(pref->stringValue), mSourceProfile, getter_AddRefs(srcSignatureFile));
      
      // now turn this into a relative pref, relative to our destination profile
      nsCOMPtr<nsILocalFile> targetProfile = do_QueryInterface(mTargetProfile);
      nsCAutoString descriptorString;
      ConvertToRelativePref(srcSignatureFile, targetProfile, descriptorString);
      NS_Free(pref->stringValue);
      pref->stringValue = ToNewCString(descriptorString);
    }
  }
  return NS_OK;
}

/**
 * the directory service is not up (and if it was, it would be for the new profile). So we need to convert
 * relative prefs from the source profile into full paths on our own. This code is copied from nsPrefBranch::GetComplexValue
 */
nsresult
nsThunderbirdProfileMigrator::ConvertFromRelativePref(const nsCString& aUTF8RelativePath,
                                                       nsILocalFile * aProfilePath, nsILocalFile **aFullPath)
{
  nsACString::const_iterator keyBegin, strEnd;
  aUTF8RelativePath.BeginReading(keyBegin);
  aUTF8RelativePath.EndReading(strEnd);    

  // The pref has the format: [fromKey]a/b/c
  if (*keyBegin++ != '[')
    return NS_ERROR_FAILURE;
  nsACString::const_iterator keyEnd(keyBegin);
  if (!FindCharInReadable(']', keyEnd, strEnd))
      return NS_ERROR_FAILURE;
  nsCAutoString key(Substring(keyBegin, keyEnd));
  NS_NewNativeLocalFile(EmptyCString(), PR_TRUE, aFullPath);
  return (*aFullPath)->SetRelativeDescriptor(aProfilePath, Substring(++keyEnd, strEnd));
}

nsresult
nsThunderbirdProfileMigrator::ConvertToRelativePref(nsILocalFile * aFulPath, nsILocalFile *aProfilePath,
                                                     nsCString& aUTF8RelativePath)
{
  nsCAutoString relDescriptor;
  nsresult rv = aFulPath->GetRelativeDescriptor(aProfilePath, relDescriptor);
  if (NS_FAILED(rv))
    return rv;

  aUTF8RelativePath.Append('[');
  aUTF8RelativePath.Append(NS_APP_USER_PROFILE_50_DIR);
  aUTF8RelativePath.Append(']');
  aUTF8RelativePath.Append(relDescriptor);
}

