/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Pierre Phaneuf <pp@ludusdesign.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*

  Outlook Express (Win32) settings

*/

#include "nsCOMPtr.h"
#include "nscore.h"
#include "nsReadableUtils.h"
#include "nsOutlookImport.h"
#include "nsIComponentManager.h"
#include "nsIServiceManager.h"
#include "nsIImportService.h"
#include "nsOutlookRegUtil.h"
#include "nsIMsgAccountManager.h"
#include "nsIMsgAccount.h"
#include "nsIImportSettings.h"
#include "nsOutlookSettings.h"
#include "nsMsgBaseCID.h"
#include "nsMsgCompCID.h"
#include "nsISmtpService.h"
#include "nsISmtpServer.h"
#include "nsOutlookStringBundle.h"
#include "OutlookDebugLog.h"
#include "nsIPop3IncomingServer.h"
#include "nsMsgI18N.h"
#include "nsMsgUtils.h"

#define DEFAULT_LOCAL_SENT_FOLDER_NAME  "Sent"

class OutlookSettings {
public:
  static HKEY FindAccountsKey( void);
  static HKEY FindOutlook2007Key( void);
  static PRBool DoImport( nsIMsgAccount **ppAccount);

  static PRBool DoIMAPServer( nsIMsgAccountManager *pMgr, HKEY hKey, char *pServerName, nsIMsgAccount **ppAccount);
  static PRBool DoPOP3Server( nsIMsgAccountManager *pMgr, HKEY hKey, char *pServerName, nsIMsgAccount **ppAccount);

  static void SetIdentities( nsIMsgAccountManager *pMgr, nsIMsgAccount *pAcc, HKEY hKey);
  static PRBool IdentityMatches( nsIMsgIdentity *pIdent, const char *pName, const char *pServer, const char *pEmail, const char *pReply, const char *pUserName);

  static void SetSmtpServer( nsIMsgAccountManager *pMgr, nsIMsgIdentity * identity, nsIMsgAccount *pAcc, char *pServer, const nsCString& pUser, HKEY hKey);
  static nsresult GetAccountName(HKEY hKey, char *defaultName, nsString &acctName);

protected:
  static nsresult generateSpecialFolderURI(nsIMsgIncomingServer * aIncomingServer, const nsString& aFolderName, nsCString& aFolderURI);
  static nsresult generateLocalFolderURI(const nsString& aFolderName,  nsCString& aFolderURI);
};


////////////////////////////////////////////////////////////////////////
nsresult nsOutlookSettings::Create(nsIImportSettings** aImport)
{
    NS_PRECONDITION(aImport != nsnull, "null ptr");
    if (! aImport)
        return NS_ERROR_NULL_POINTER;

    *aImport = new nsOutlookSettings();
    if (! *aImport)
        return NS_ERROR_OUT_OF_MEMORY;

    NS_ADDREF(*aImport);
    return NS_OK;
}

nsOutlookSettings::nsOutlookSettings()
{
}

nsOutlookSettings::~nsOutlookSettings()
{
}

NS_IMPL_ISUPPORTS1(nsOutlookSettings, nsIImportSettings)

NS_IMETHODIMP nsOutlookSettings::AutoLocate(PRUnichar **description, nsIFile **location, PRBool *_retval)
{
    NS_PRECONDITION(description != nsnull, "null ptr");
    NS_PRECONDITION(_retval != nsnull, "null ptr");
  if (!description || !_retval)
    return( NS_ERROR_NULL_POINTER);

  *description = nsOutlookStringBundle::GetStringByID( OUTLOOKIMPORT_NAME);
  *_retval = PR_FALSE;

  if (location)
    *location = nsnull;

  // look for the registry key for the accounts
  HKEY key = OutlookSettings::FindAccountsKey();
  if (key != nsnull) {
    *_retval = PR_TRUE;
    ::RegCloseKey( key);
  }

  return( NS_OK);
}

NS_IMETHODIMP nsOutlookSettings::SetLocation(nsIFile *location)
{
  return( NS_OK);
}

NS_IMETHODIMP nsOutlookSettings::Import(nsIMsgAccount **localMailAccount, PRBool *_retval)
{
  NS_PRECONDITION( _retval != nsnull, "null ptr");

  if (OutlookSettings::DoImport( localMailAccount)) {
    *_retval = PR_TRUE;
    IMPORT_LOG0( "Settings import appears successful\n");
  }
  else {
    *_retval = PR_FALSE;
    IMPORT_LOG0( "Settings import returned FALSE\n");
  }

  return( NS_OK);
}

HKEY OutlookSettings::FindOutlook2007Key (void)
{
  // Outlook 2002 and above write to a different location
  HKEY sKey;
  nsCString defaultProfile("Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows Messaging Subsystem\\Profiles");
  if (::RegOpenKeyEx( HKEY_CURRENT_USER, defaultProfile.get(), 0, KEY_QUERY_VALUE, &sKey) == ERROR_SUCCESS) 
  {
    // Find the default profile name (usually it's Outlook)
    char * defaultPathBuffer = (char *) nsOutlookRegUtil::GetValueBytes( sKey, "DefaultProfile");
    if (defaultPathBuffer)
    {
      defaultProfile.AppendLiteral("\\");
      // append the profile name
      defaultProfile.Append(defaultPathBuffer);
      // append the Outlook guid
      defaultProfile.AppendLiteral("\\9375CFF0413111d3B88A00104B2A6676");
      if (::RegOpenKeyEx( HKEY_CURRENT_USER, defaultProfile.get(), 0, KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS, &sKey) == ERROR_SUCCESS)
        return( sKey);
    }  
  }
  return( nsnull);
}

HKEY OutlookSettings::FindAccountsKey( void)
{
  HKEY  sKey;

  sKey = FindOutlook2007Key();
  if (sKey)
    return( sKey);
  if (::RegOpenKeyEx( HKEY_CURRENT_USER, "Software\\Microsoft\\Office\\Outlook\\OMI Account Manager\\Accounts", 0, KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS, &sKey) == ERROR_SUCCESS) {
    return( sKey);
  }
  if (::RegOpenKeyEx( HKEY_CURRENT_USER, "Software\\Microsoft\\Office\\8.0\\Outlook\\OMI Account Manager\\Accounts", 0, KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS, &sKey) == ERROR_SUCCESS) {
    return( sKey);
  }

  return( nsnull);
}

PRBool OutlookSettings::DoImport( nsIMsgAccount **ppAccount)
{
  HKEY  hKey = FindAccountsKey();
  if (hKey == nsnull) {
    IMPORT_LOG0( "*** Error finding Outlook registry account keys\n");
    return( PR_FALSE);
  }

  nsresult  rv;

  nsCOMPtr<nsIMsgAccountManager> accMgr =
           do_GetService(NS_MSGACCOUNTMANAGER_CONTRACTID, &rv);
    if (NS_FAILED(rv)) {
    IMPORT_LOG0( "*** Failed to create a account manager!\n");
    return( PR_FALSE);
  }

  HKEY    subKey = NULL;
  nsCString  defMailName;
  PRBool binaryKeys;
  // Outlook 2007 writes to a different location
  subKey = FindOutlook2007Key();
  if (!subKey)
    if (::RegOpenKeyEx( HKEY_CURRENT_USER, "Software\\Microsoft\\Office\\Outlook\\OMI Account Manager", 0, KEY_QUERY_VALUE, &subKey) != ERROR_SUCCESS)
      if (::RegOpenKeyEx( HKEY_CURRENT_USER, "Software\\Microsoft\\Office\\8.0\\Outlook\\OMI Account Manager", 0, KEY_QUERY_VALUE, &subKey) != ERROR_SUCCESS)
        subKey = NULL;

  if (subKey != NULL) {
    // First let's get the default mail account key name
    BYTE *  pBytes = nsOutlookRegUtil::GetValueBytes( subKey, "Default Mail Account");
    ::RegCloseKey( subKey);
    if (pBytes) {
      defMailName = (const char *)pBytes;
      nsOutlookRegUtil::FreeValueBytes( pBytes);
    }
  }

  // Iterate the accounts looking for POP3 & IMAP accounts...
  // Ignore LDAP & NNTP for now!
  DWORD    index = 0;
  DWORD    numChars;
  TCHAR    keyName[256];
  FILETIME  modTime;
  LONG    result = ERROR_SUCCESS;
  BYTE *    pBytes;
  int      popCount = 0;
  int      accounts = 0;
  nsCString  keyComp;

  while (result == ERROR_SUCCESS) {
    numChars = 256;
    result = ::RegEnumKeyEx( hKey, index, keyName, &numChars, NULL, NULL, NULL, &modTime);
    index++;
    if (result == ERROR_SUCCESS) {
      if (::RegOpenKeyEx( hKey, keyName, 0, KEY_QUERY_VALUE, &subKey) == ERROR_SUCCESS) {
        // Get the values for this account.
        IMPORT_LOG1( "Opened Outlook account: %s\n", (char *)keyName);

        nsIMsgAccount  *anAccount = nsnull;
        pBytes = nsOutlookRegUtil::GetValueBytes( subKey, "IMAP Server");
        if (pBytes) {
          if (DoIMAPServer( accMgr, subKey, (char *)pBytes, &anAccount))
            accounts++;
          nsOutlookRegUtil::FreeValueBytes( pBytes);
        }

        pBytes = nsOutlookRegUtil::GetValueBytes( subKey, "POP3 Server");
        if (pBytes) {
          if (popCount == 0) {
            if (DoPOP3Server( accMgr, subKey, (char *)pBytes, &anAccount)) {
              popCount++;
              accounts++;
              if (ppAccount && anAccount) {
                *ppAccount = anAccount;
                NS_ADDREF( anAccount);
              }
            }
          }
          else {
            if (DoPOP3Server( accMgr, subKey, (char *)pBytes, &anAccount)) {
              popCount++;
              accounts++;
              // If we created a mail account, get rid of it since
              // we have 2 POP accounts!
              if (ppAccount && *ppAccount) {
                NS_RELEASE( *ppAccount);
                *ppAccount = nsnull;
              }
            }
          }
          nsOutlookRegUtil::FreeValueBytes( pBytes);
        }

        if (anAccount) {
          // Is this the default account?
          keyComp = keyName;
          if (keyComp.Equals( defMailName)) {
            accMgr->SetDefaultAccount( anAccount);
          }
          NS_RELEASE( anAccount);
        }

        ::RegCloseKey( subKey);
      }
    }
  }

  // Now save the new acct info to pref file.
  rv = accMgr->SaveAccountInfo();
  NS_ASSERTION(NS_SUCCEEDED(rv), "Can't save account info to pref file");

  return( accounts != 0);
}

nsresult OutlookSettings::GetAccountName(HKEY hKey, char *defaultName, nsString &acctName)
{
  BYTE *pAccName = nsOutlookRegUtil::GetValueBytes( hKey, "Account Name");
  nsresult rv = NS_OK;
  if (pAccName) {
     rv = nsMsgI18NConvertToUnicode(nsMsgI18NFileSystemCharset(),
                                    nsDependentCString( (char *) pAccName), acctName);
    nsOutlookRegUtil::FreeValueBytes( pAccName);
  }
  else
    acctName.AssignASCII(defaultName);
  return rv;
}

PRBool OutlookSettings::DoIMAPServer( nsIMsgAccountManager *pMgr, HKEY hKey, char *pServerName, nsIMsgAccount **ppAccount)
{
  if (ppAccount)
    *ppAccount = nsnull;

  BYTE *pBytes;
  pBytes = nsOutlookRegUtil::GetValueBytes( hKey, "IMAP User Name");
  if (!pBytes)
  {
    // Outlook 2007 uses IMAP User
    pBytes = nsOutlookRegUtil::GetValueBytes( hKey, "IMAP User");
    if (!pBytes)
      return( PR_FALSE);
  }
  PRBool  result = PR_FALSE;
  
  // I now have a user name/server name pair, find out if it already exists?
  nsCOMPtr<nsIMsgIncomingServer> in;
  nsresult rv = pMgr->FindServer( nsDependentCString((const char *)pBytes), nsDependentCString(pServerName), NS_LITERAL_CSTRING("imap"), getter_AddRefs( in));
  if (NS_FAILED( rv) || (in == nsnull)) {
    // Create the incoming server and an account for it?
    rv = pMgr->CreateIncomingServer( nsDependentCString((const char *)pBytes), nsDependentCString(pServerName), NS_LITERAL_CSTRING("imap"), getter_AddRefs( in));
    if (NS_SUCCEEDED( rv) && in) {
      rv = in->SetType(NS_LITERAL_CSTRING("imap"));
      IMPORT_LOG2( "Created IMAP server named: %s, userName: %s\n", pServerName, (char *)pBytes);

      BYTE * pPort = nsOutlookRegUtil::GetValueBytes(hKey, "IMAP Port");
      PRInt32 port = 0;
      if (pPort)
      {
        port = *(PRInt32 *) pPort; // don't set this immediately
        nsOutlookRegUtil::FreeValueBytes(pPort);
      }

      // Set the SSL connection
      BYTE * pSecureConnection = nsOutlookRegUtil::GetValueBytes(hKey, "IMAP Use SSL");
      if (pSecureConnection)
      {
        PRInt32 secureConnection = (*(PRInt32 *) pSecureConnection);
        IMPORT_LOG2( "IMAP server named: %s, has IMAP Use SSL value of %d\n", pServerName, secureConnection);
        if (secureConnection == 1)
        {
          in->SetSocketType(nsMsgSocketType::SSL);
          if (!port)
            port = 993; // the default ssl imap port
        }
        else if (secureConnection == 2) // TLS
          in->SetSocketType(nsMsgSocketType::alwaysSTARTTLS);
        else if (secureConnection == 3) // Auto - we don't have an equivalent...try tls?
          in->SetSocketType(nsMsgSocketType::alwaysSTARTTLS);
        nsOutlookRegUtil::FreeValueBytes(pSecureConnection);
      }

      if (port)
        in->SetPort(port);

      BYTE * pBytesTemp = nsOutlookRegUtil::GetValueBytes(hKey, "IMAP Use SPA");
      PRInt32 authMethod;
      if (pBytesTemp)
      {
        PRBool secAuth = *(PRBool *)pBytesTemp;
        nsOutlookRegUtil::FreeValueBytes(pBytesTemp);
        IMPORT_LOG2( "IMAP server named: %s, has IMAP Use SPA of %s\n", pServerName, secAuth ? "true" : "false");
        authMethod = secAuth ? nsMsgAuthMethod::secure
            : nsMsgAuthMethod::passwordCleartext;
      }
      else
        authMethod = nsMsgAuthMethod::passwordCleartext;
      in->SetAuthMethod(authMethod);
      
      nsString prettyName;
      if (NS_SUCCEEDED(GetAccountName(hKey, pServerName, prettyName)))
        rv = in->SetPrettyName( prettyName);

      // We have a server, create an account.
      nsCOMPtr<nsIMsgAccount>  account;
      rv = pMgr->CreateAccount( getter_AddRefs( account));
      if (NS_SUCCEEDED( rv) && account) {
        rv = account->SetIncomingServer( in);

        IMPORT_LOG0( "Created an account and set the IMAP server as the incoming server\n");

        // Fiddle with the identities
        SetIdentities( pMgr, account, hKey);
        result = PR_TRUE;
        if (ppAccount)
          account->QueryInterface( NS_GET_IID(nsIMsgAccount), (void **)ppAccount);
      }
    }
  }
  else
    result = PR_TRUE;

  nsOutlookRegUtil::FreeValueBytes( pBytes);

  return( result);
}

PRBool OutlookSettings::DoPOP3Server( nsIMsgAccountManager *pMgr, HKEY hKey, char *pServerName, nsIMsgAccount **ppAccount)
{
  if (ppAccount)
    *ppAccount = nsnull;

  BYTE *pBytes;
  pBytes = nsOutlookRegUtil::GetValueBytes( hKey, "POP3 User Name");
  if (!pBytes)
  {
    pBytes = nsOutlookRegUtil::GetValueBytes( hKey, "POP3 User"); // Outlook 2007 and above
    if (!pBytes)
      return( PR_FALSE);
  }

  PRBool result = PR_FALSE;

  // I now have a user name/server name pair, find out if it already exists?
  nsCOMPtr<nsIMsgIncomingServer> in;
  nsresult rv = pMgr->FindServer( nsDependentCString((const char *)pBytes), nsDependentCString(pServerName), NS_LITERAL_CSTRING("pop3"), getter_AddRefs( in));
  if (NS_FAILED( rv) || (in == nsnull)) {
    // Create the incoming server and an account for it?
    rv = pMgr->CreateIncomingServer(nsDependentCString((const char *)pBytes), nsDependentCString(pServerName), NS_LITERAL_CSTRING("pop3"), getter_AddRefs( in));
    if (NS_SUCCEEDED( rv) && in) {
      rv = in->SetType(NS_LITERAL_CSTRING("pop3"));

      BYTE * pPort = nsOutlookRegUtil::GetValueBytes(hKey, "POP3 Port");
      PRInt32 port = 0;
      if (pPort)
      {
        port = *(PRInt32 *) pPort; // don't set the port yet
        IMPORT_LOG2( "POP3 server named: %s, has POP3 Port value of %d\n", pServerName, port);
        nsOutlookRegUtil::FreeValueBytes(pPort);
      }

      // Set the SSL connection
      BYTE * pSecureConnection = nsOutlookRegUtil::GetValueBytes(hKey, "POP3 Use SSL");
      if (pSecureConnection)
      {
        PRInt32 secureConnection = (*(PRInt32 *) pSecureConnection);
        IMPORT_LOG2( "POP3 server named: %s, has POP3 Use SSL value of %d\n", pServerName, secureConnection);
        if (secureConnection == 1)
        {
          in->SetSocketType(nsMsgSocketType::SSL);
          if (!port)
            port = 995; // the default SSL port for POP3
        }
        else if (secureConnection == 2) // TLS
          in->SetSocketType(nsMsgSocketType::alwaysSTARTTLS);
        else if (secureConnection == 3) // Auto - we don't have an equivalent...try tls?
          in->SetSocketType(nsMsgSocketType::alwaysSTARTTLS);
        nsOutlookRegUtil::FreeValueBytes(pSecureConnection);
      }

      if (port)
        in->SetPort(port);

      BYTE * pBytesTemp = nsOutlookRegUtil::GetValueBytes(hKey, "POP3 Use SPA");
      PRInt32 authMethod;
      if (pBytesTemp)
      {
        PRBool secAuth = *(PRBool *)pBytesTemp;
        nsOutlookRegUtil::FreeValueBytes(pBytesTemp);
        IMPORT_LOG2( "POP3 server named: %s, has POP3 Use SPA of %s\n", pServerName, secAuth ? "true" : "false");
        authMethod = secAuth ? nsMsgAuthMethod::secure
            : nsMsgAuthMethod::passwordCleartext;
      }
      else
        authMethod = nsMsgAuthMethod::passwordCleartext;
      in->SetAuthMethod(authMethod);

        IMPORT_LOG2( "Created POP3 server named: %s, userName: %s\n", pServerName, (char *)pBytes);

        nsString prettyName;
        if (NS_SUCCEEDED(GetAccountName(hKey, pServerName, prettyName)))
          rv = in->SetPrettyName( prettyName);
      // We have a server, create an account.
      nsCOMPtr<nsIMsgAccount>  account;
      rv = pMgr->CreateAccount( getter_AddRefs( account));
      if (NS_SUCCEEDED( rv) && account) {
        rv = account->SetIncomingServer( in);

      IMPORT_LOG0( "Created a new account and set the incoming server to the POP3 server.\n");

        nsCOMPtr<nsIPop3IncomingServer> pop3Server = do_QueryInterface(in, &rv);
        NS_ENSURE_SUCCESS(rv,rv);
        BYTE *pLeaveOnServer = nsOutlookRegUtil::GetValueBytes( hKey, "Leave Mail On Server");
        if (!pLeaveOnServer)
          pLeaveOnServer = nsOutlookRegUtil::GetValueBytes( hKey, "Leave On Server"); // Outlook 2007 and above
        if (pLeaveOnServer)
        {
          pop3Server->SetLeaveMessagesOnServer(*pLeaveOnServer == 1 ? PR_TRUE : PR_FALSE);
          nsOutlookRegUtil::FreeValueBytes(pLeaveOnServer);
        }

        // Fiddle with the identities
        SetIdentities( pMgr, account, hKey);
        result = PR_TRUE;
        if (ppAccount)
          account->QueryInterface( NS_GET_IID(nsIMsgAccount), (void **)ppAccount);
      }
    }
  }
  else
    result = PR_TRUE;

  nsOutlookRegUtil::FreeValueBytes( pBytes);

  return( result);
}

PRBool OutlookSettings::IdentityMatches( nsIMsgIdentity *pIdent, const char *pName, const char *pServer, const char *pEmail, const char *pReply, const char *pUserName)
{
  if (!pIdent)
    return( PR_FALSE);

  char *  pIName = nsnull;
  nsCString pIEmail;
  nsCString pIReply;

  PRBool  result = PR_TRUE;

  // The test here is:
  // If the smtp host is the same
  //  and the email address is the same (if it is supplied)
  //  and the reply to address is the same (if it is supplied)
  //  then we match regardless of the full name.


  nsresult rv;
  rv = pIdent->GetEmail(pIEmail);
  rv = pIdent->GetReplyTo(pIReply);


  // for now, if it's the same server and reply to and email then it matches
  if (pReply && !pIReply.Equals(pReply, nsCaseInsensitiveCStringComparator()))
    result = PR_FALSE;
  if (pEmail && !pIEmail.Equals(pEmail, nsCaseInsensitiveCStringComparator()))
    result = PR_FALSE;

  return( result);
}

void OutlookSettings::SetIdentities( nsIMsgAccountManager *pMgr, nsIMsgAccount *pAcc, HKEY hKey)
{
  // Get the relevant information for an identity
  char *pName = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "SMTP Display Name");
  if (!pName)
    pName = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "Display Name");
  char *pServer = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "SMTP Server");
  char *pEmail = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "SMTP Email Address");
  if (!pEmail)
    pEmail = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "Email");  // Outlook 2007 key
  char *pReply = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "SMTP Reply To Email Address");
  if (!pReply)
    pReply = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "Reply Email"); // Outlook 2007 key
  nsCString userName;
  userName.Adopt((char *)nsOutlookRegUtil::GetValueBytes( hKey, "SMTP User Name"));
  char *pOrgName = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "SMTP Organization Name");
  if (!pOrgName)
    pOrgName = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "Organization"); // Outlook 2007 key
  char *pImapSentItemsFlag = (char *)nsOutlookRegUtil::GetValueBytes( hKey, "IMAP Sentitems flag");
  PRInt32 sentItemsFlag = 0;
  if (pImapSentItemsFlag)  
   sentItemsFlag = *(PRInt32 *) pImapSentItemsFlag;

  nsresult rv;

  nsCOMPtr<nsIMsgIdentity>  id;
  if (pEmail && pName && pServer) {
    nsCOMPtr <nsIMsgIncomingServer>  incomingServer;
    pAcc->GetIncomingServer(getter_AddRefs( incomingServer));

    // The default identity, nor any other identities matched,
    // create a new one and add it to the account.
    rv = pMgr->CreateIdentity( getter_AddRefs( id));
    if (id) {
      nsAutoString name, organization;
      rv = nsMsgI18NConvertToUnicode(nsMsgI18NFileSystemCharset(),
        nsDependentCString(pName), name);
      if (NS_SUCCEEDED(rv))
      {
        id->SetFullName(name);
        id->SetIdentityName(name);
        IMPORT_LOG1( "Created identity  named: %s\n", pName);
      }

      if (pOrgName) {
        rv = nsMsgI18NConvertToUnicode(nsMsgI18NFileSystemCharset(),
          nsDependentCString(pOrgName), organization);
        if (NS_SUCCEEDED(rv))
        {
          id->SetOrganization(organization);
          IMPORT_LOG2( "identity named: %s with organization %s\n", pName, pOrgName);
        }
      }

      id->SetEmail(nsDependentCString(pEmail));
      IMPORT_LOG2( "identity named: %s with email %s\n", pName, pEmail);
      if (pReply)
      {
        id->SetReplyTo(nsDependentCString(pReply));
        IMPORT_LOG2( "identity named: %s with email %s\n", pName, pReply);
      }

      nsCString folderURI;
      
      IMPORT_LOG1("imap sent items flag set to: %d", sentItemsFlag);
      // use our default drafts folder
      if (incomingServer)
      {
        // currently not supporting migratin for folders other than the default so just set it to Sent for the account
        generateSpecialFolderURI(incomingServer, NS_LITERAL_STRING("Sent"), folderURI);
        id->SetFccFolder(folderURI);
        id->SetFccFolderPickerMode(NS_LITERAL_CSTRING("0"));
        IMPORT_LOG1("fcc folder set to: %s", folderURI.get());
      
        generateSpecialFolderURI(incomingServer, NS_LITERAL_STRING("Drafts"), folderURI);
        id->SetDraftFolder(folderURI);
        id->SetDraftsFolderPickerMode(NS_LITERAL_CSTRING("0"));
        IMPORT_LOG1("drafts folder set to: %s", folderURI.get());

        // use our default templates folder
        generateSpecialFolderURI(incomingServer, NS_LITERAL_STRING("Templates"), folderURI);
        id->SetStationeryFolder(folderURI);
        id->SetTmplFolderPickerMode(NS_LITERAL_CSTRING("0"));
        IMPORT_LOG1("template folder set to: %s", folderURI.get());
      }

      pAcc->AddIdentity( id);

      IMPORT_LOG0( "Created identity and added to the account\n");
      IMPORT_LOG1( "\tname: %s\n", pName);
      IMPORT_LOG1( "\temail: %s\n", pEmail);
    }
  }

  if (userName.IsEmpty()) {
    nsCOMPtr <nsIMsgIncomingServer>  incomingServer;
    rv = pAcc->GetIncomingServer(getter_AddRefs( incomingServer));
    if (NS_SUCCEEDED(rv) && incomingServer)
      rv = incomingServer->GetUsername(userName);
    NS_ASSERTION(NS_SUCCEEDED(rv), "Unable to get UserName from incomingServer");
  }

  SetSmtpServer( pMgr, id, pAcc, pServer, userName, hKey);

  nsOutlookRegUtil::FreeValueBytes( (BYTE *)pName);
  nsOutlookRegUtil::FreeValueBytes( (BYTE *)pServer);
  nsOutlookRegUtil::FreeValueBytes( (BYTE *)pEmail);
  nsOutlookRegUtil::FreeValueBytes( (BYTE *)pReply);
  nsOutlookRegUtil::FreeValueBytes( (BYTE *)pImapSentItemsFlag);
}

void OutlookSettings::SetSmtpServer(nsIMsgAccountManager *pMgr, nsIMsgIdentity * identity, nsIMsgAccount *pAcc, char *pServer,
                                    const nsCString& user, HKEY hKey)
{
  nsresult rv;
  nsCOMPtr<nsISmtpService> smtpService(do_GetService(NS_SMTPSERVICE_CONTRACTID, &rv));
  if (NS_SUCCEEDED(rv) && smtpService) {
    nsCString hostName(pServer);
    nsCString userName(user);
    BYTE * pBytes = nsOutlookRegUtil::GetValueBytes( hKey, "SMTP Server");
    if (pBytes)
    {
      // Outlook 2007 and above supports a SMTP Server field for the host name. use it over pServer if available
      hostName = (char *) pBytes;
      IMPORT_LOG1( "overriding host name with: %s\n", (char *) pBytes);
      nsOutlookRegUtil::FreeValueBytes(pBytes);
    }

    pBytes = nsOutlookRegUtil::GetValueBytes( hKey, "SMTP User");
    if (pBytes)
    {
      // Outlook 2007 and above supports a SMTP User field for the host name. use it over pServer if available
      userName = (char *) pBytes;
      IMPORT_LOG1( "overriding user name with: %s\n", (char *) pBytes);
      nsOutlookRegUtil::FreeValueBytes(pBytes);
    }

    nsCOMPtr<nsISmtpServer> foundServer;
    rv = smtpService->FindServer( userName.get(), hostName.get(), getter_AddRefs( foundServer));
    if (NS_SUCCEEDED( rv) && foundServer) {
      IMPORT_LOG1( "SMTP server already exists: %s\n", hostName.get());
      return;
    }

    nsCOMPtr<nsISmtpServer> smtpServer;
    rv = smtpService->CreateSmtpServer( getter_AddRefs( smtpServer));
    if (NS_SUCCEEDED( rv) && smtpServer) {
      smtpServer->SetHostname(hostName);
      IMPORT_LOG1( "created smtp server with host name: %s\n", hostName.get());
      
      if (!userName.IsEmpty())
        smtpServer->SetUsername(userName);
      IMPORT_LOG2( "smtp server with host name: %s and user name:%s\n", hostName.get(), userName.get());

      pBytes = nsOutlookRegUtil::GetValueBytes(hKey, "SMTP Port");
      PRInt32 port = 0;
      if (pBytes)
      {
        port = *(PRInt32 *) pBytes;
        IMPORT_LOG2( "smtp server with host name: %s and port:%d\n", hostName.get(), *(PRInt32 *) pBytes);
        nsOutlookRegUtil::FreeValueBytes(pBytes);
      }

      // Set the SSL connection
      BYTE * pSecureConnection = nsOutlookRegUtil::GetValueBytes(hKey, "SMTP Use SSL");
      if (!pSecureConnection)
        pSecureConnection = nsOutlookRegUtil::GetValueBytes(hKey, "SMTP Secure Connection"); // my version of outlook 2007 uses this key not SMTP Use SSL
      if (pSecureConnection)
      {
        PRInt32 secureConnection = (*(PRInt32 *) pSecureConnection);
        IMPORT_LOG2( "smtp server with host name: %s and SMTP Use SSL:%d\n", hostName.get(), secureConnection);
        if (secureConnection == 1)
        {
          smtpServer->SetSocketType(nsMsgSocketType::SSL);
          if (!port)
            port = 465;
        }
        else if (secureConnection == 2) // TLS
          smtpServer->SetSocketType(nsMsgSocketType::alwaysSTARTTLS);
        else if (secureConnection == 3) // Auto - we don't have an equivalent...try tls?
          smtpServer->SetSocketType(nsMsgSocketType::alwaysSTARTTLS);
        nsOutlookRegUtil::FreeValueBytes(pSecureConnection);
      }

      if (port)
        smtpServer->SetPort(port);

      pBytes = nsOutlookRegUtil::GetValueBytes(hKey, "SMTP Use Auth");
      if (pBytes)
      {
        // SMTP Use Auth is the boolean for whether the SMTP server needs authentication
        PRBool useAuthentication = *(PRBool *)pBytes;
        IMPORT_LOG2( "smtp server with host name: %s and SMTP Use Auth:%s\n", hostName.get(), useAuthentication ? "true" : "false");
        nsOutlookRegUtil::FreeValueBytes(pBytes);
        PRInt32 authMethod = nsMsgAuthMethod::none;
        if (useAuthentication)
        {
          pBytes = nsOutlookRegUtil::GetValueBytes(hKey, "SMTP Use SPA");
          if (pBytes)
          {
            PRBool secAuth = *(PRBool *)pBytes;
            IMPORT_LOG2( "smtp server with host name: %s and SMTP Use SPA:%s\n", hostName.get(), secAuth ? "true" : "false");
            nsOutlookRegUtil::FreeValueBytes(pBytes);
            authMethod = secAuth ? nsMsgAuthMethod::secure
                : nsMsgAuthMethod::passwordCleartext;
          }
          else
            authMethod = nsMsgAuthMethod::passwordCleartext;
        }
        smtpServer->SetAuthMethod(authMethod);
      }

       // associate the SMTP server with the identity
      nsCString smtpServerKey;
      smtpServer->GetKey(getter_Copies(smtpServerKey));
      if (identity && !smtpServerKey.IsEmpty())
      {
        identity->SetSmtpServerKey(smtpServerKey);
        // don't allow gmail SMTP servers to copy to the fcc folder
        if (hostName.LowerCaseEqualsLiteral("smtp.gmail.com") || hostName.LowerCaseEqualsLiteral("smtp.googlemail.com"))
          identity->SetDoFcc(PR_FALSE);
      }
      IMPORT_LOG1( "Created new SMTP server: %s\n", pServer);
    }
  }
}

/**
 * when an imap account is setup to NOT use an online folder for Sent, Drafts, Junk
 * we want to create a folderURI under Local folders where the folder name has the following format: 
 * Sent Messages (account name), Drafts (account name), etc.
 */
nsresult OutlookSettings::generateLocalFolderURI(const nsString& aFolderName,  nsCString& aFolderURI)
{
  nsresult rv;
  nsCOMPtr<nsIMsgAccountManager> accMgr = do_GetService(NS_MSGACCOUNTMANAGER_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIMsgIncomingServer> localFoldersServer;
  rv = accMgr->GetLocalFoldersServer(getter_AddRefs(localFoldersServer));
  // if Local Folders does not exist already, create it
  if (NS_FAILED(rv) || !localFoldersServer)
  {
    rv = accMgr->CreateLocalMailAccount();
    NS_ENSURE_SUCCESS(rv, rv);
    rv = accMgr->GetLocalFoldersServer(getter_AddRefs(localFoldersServer));
  }

  NS_ENSURE_SUCCESS(rv, rv);

  // the folder name is always of the form: "Folder Name"
  return generateSpecialFolderURI(localFoldersServer, aFolderName, aFolderURI);
}

/**
 * helper method used to take the root folder uri for an incoming server and append aFolderName to it thus generating the URI
 * for a special folder which can then be set as a pref.
 */
nsresult OutlookSettings::generateSpecialFolderURI(nsIMsgIncomingServer * aIncomingServer, const nsString& aFolderName, nsCString& aFolderURI)
{
  nsCOMPtr<nsIMsgFolder> rootMsgFolder;
  nsresult rv = aIncomingServer->GetRootFolder(getter_AddRefs(rootMsgFolder));
  NS_ENSURE_SUCCESS(rv, rv);

  rv = rootMsgFolder->GetURI(aFolderURI);
  NS_ENSURE_SUCCESS(rv, rv);

  aFolderURI += "/";
  

  // folder name needs to be escaped before being appended to a URI
  nsCString escapedName;
  rv = NS_MsgEscapeEncodeURLPath(aFolderName, escapedName);
  NS_ENSURE_SUCCESS(rv, rv);
  aFolderURI += escapedName;  
  return NS_OK;
}
